/* camera.c - Rotinas da Biblioteca Grafica de Implementacao da
              Camera Virtual */

#include <math.h>
#include <stdio.h>
#include <f.h>
#include <camera.h>

static ponto3d _posicao = {0.0, 0.0, 0.0},  /* posicao da camera */
               _alvo = {0.0, 0.0, 0.0},     /* ponto para onde o observador
                                               esta olhando */
               _direcao = {0.0, 0.0, -1.0}, /* vetor direcao de visao */
               _normal = {0.0, 0.0, -1.0},  /* vetor normal ao plano de
                                               projecao */
	            _vertical = {0.0, 1.0, 0.0}; /* vetor orientacao da
                                               camera (observador) */

static ponto2d O = {0.0, 0.0},
               f, g; /* variaveis auxiliares para o calculo
                        do ponto na tela */

static int alvo_ativo = FALSE,
           tipo = PARALELA,
           esqt = 0, inft = 200,  /*  coordenadas da janela da tela */
           dirt = 320, supt = 0;  /*  (esquerda inferior e direita superior) */

static float _foco = 1,    /* posicao do foco da camera, ponto onde
                              o plano de projecao corta o eixo z */
             _alfa, _beta, /* angulos de abertura vertical e horizontal
                              medidos pela reta que vai de posicao ate os
                              extremos da janela do mundo */
             _epsilon=0.1, /* valor pequeno para incremento */
             _perto=1.0,       /* plano de corte, distancia minima de visao */
             esqj=-1.0, infj=-1.0, /* coordenadas da janela do mundo */
             dirj=1.0, supj=1.0; /* (esquerda inferior e direita superior) */

/* ponteiro para a funcao de projecao escolhida (paralela ou perspectiva) */
void (*projecao) () = projecao_paralela;

static matriz3x3 _mundo_obs,  /* base do observador */
                 _obs_mundo;  /* base do observador transposta */


/*-------------------------------------------------------------*\
 | void janela (float e, float i, float d, float s)            |
 | Define a janela do mundo no plano de projecao.              |
 | Os limites da janela sao:                                   |
 | e: esquerdo.                                                |
 | i: inferior.                                                |
 | d: direito.                                                 |
 | s: superior.                                                |
\*-------------------------------------------------------------*/
void janela (float e, float i, float d, float s)
{
  esqj = e + O.x;
  infj = i + O.y;
  dirj = d + O.x;
  supj = s + O.y;
}


void mostra_camera()
{
 int i,j;
 
 printf("Base do mundo_obs\n");
 for(i=0; i<3; i++){
   for (j=0; j<3; j++)
     fprintf(stdout, "%3.3lf ", _mundo_obs[i][j]);
   printf("\n");
 }   
 printf("Base do obs_mundo\n");
 for(i=0; i<3; i++){
   for (j=0; j<3; j++)
     fprintf(stdout, "%3.3lf ", _obs_mundo[i][j]);
   printf("\n");
  }
  printf("\n");
}

/*-------------------------------------------------------------*\
 | void posicao (ponto3d *p).                                  |
 | Define a posicao do observador(camera).                     |
\*-------------------------------------------------------------*/
void posicao (p)
ponto3d *p;
{
  _posicao.x = p->x;
  _posicao.y = p->y;
  _posicao.z = p->z;
}


/*-------------------------------------------------------------*\
 | void alvo (ponto3d *p)                                      |
 | Define a posicao do alvo.                                   |
 | Nota: alvo e' o ponto para o qual o observador esta olhando.|
\*-------------------------------------------------------------*/
void alvo (p)
ponto3d *p;
{
  _alvo.x = p->x;
  _alvo.y = p->y;
  _alvo.z = p->z;
  subtracao (&_alvo, &_posicao, &_direcao);

  _normal.x = _direcao.x;
  _normal.y = _direcao.y;
  _normal.z = _direcao.z;

  alvo_ativo = TRUE;

  calcula_base (&_direcao, &_vertical);
}

/*-------------------------------------------------------------*\
 | void direcao (ponto3d *p)                                   |
 | Define a direcao de visao, ou direcao da camera.            |
\*-------------------------------------------------------------*/
void direcao (p)
ponto3d *p;
{
  _normal.x = _direcao.x = p->x;
  _normal.y = _direcao.y = p->y;
  _normal.z = _direcao.z = p->z;

  alvo_ativo = FALSE;

  O.x = 0.0; /* normal == direcao, logo a direcao passa pela */
  O.y = 0.0; /* origem do plano de projecao.                 */

  calcula_base (&_direcao, &_vertical);
}

/*-------------------------------------------------------------*\
 | void vertical (ponto3d *p)                                  |
 | Define vetor verticao, que indica se o observador esta em   |
 | "pe'", "deitado", "de ponta cabeca", etc.                   |
\*-------------------------------------------------------------*/
void vertical (p)
ponto3d *p;
{
  _vertical.x = p->x;
  _vertical.y = p->y;
  _vertical.z = p->z;

  calcula_base (&_direcao, &_vertical);
}

/*-------------------------------------------------------------*\
 | void foco (float d)                                         |
 | Define o foco, que e' a distancia entre o observador e o    |
 | plano de projecao.                                          |
\*-------------------------------------------------------------*/
void foco (float d)
{
  _foco = -d;
}

/*-------------------------------------------------------------*\
 | void abertura (float a, float b)                            |
 | Define o angulo de abertura horizontal e vertical da camera.|
 | Altera com isso o tamanho da janela.                        |
\*-------------------------------------------------------------*/
void abertura (float a, float b)
{
  float l = _foco * tan (a),
        m = _foco * tan (b);

  _alfa = a;
  _beta = b;

  janela (-m, -l, m, l);
}

/*-------------------------------------------------------------*\
 | void tipo_projecao (int t)                                  |
 | Define o tipo de projecao, perspectiva ou paralela.         |
\*-------------------------------------------------------------*/
void tipo_projecao (t)
int t;
{
  tipo = t;

  projecao = (tipo==PERSPECTIVA) ? projecao_perspectiva : projecao_paralela;
}

/*-------------------------------------------------------------*\
 | void normal (ponto3d *p)                                    |
 | Define a normal ao plano de projecao.                       |
 | Se a normal for diferente da direcao, entao a origem do pla-|
 | no de projecao sera dirferente de (0,0).                    |
\*-------------------------------------------------------------*/
void normal (p)
ponto3d *p;
{
  ponto3d dobs;
  float t;

  _normal.x = p->x;
  _normal.y = p->y;
  _normal.z = p->z;

  alvo_ativo = FALSE;

  vetorXmatriz (&_direcao, _mundo_obs, &dobs);

  /* calcula a origem do plano de projecao. */
  t = -_foco / dobs.z;
  O.x = t * dobs.x;
  O.y = t * dobs.y;

  calcula_base (&_normal, &_vertical);
}

/*-------------------------------------------------------------*\
 | void base (ponto3d *p, ponto3d *q, ponto3d *r)              |
 | Define uma nova base do observador.                         |
 | Nota: com esta funcao pode ser definida uma base arbitraria |
 | qualquer, o que pode causar distorcoes na cena.             |
\*-------------------------------------------------------------*/
void base (p, q, r)
ponto3d *p;
ponto3d *q;
ponto3d *r;
{

  monta_base (p, q, r);
 
  O.x = 0.0;
  O.y = 0.0;

  alvo_ativo = FALSE;
  
}
/*-------------------------------------------------------------*\
 | void tela (int e, int i, int d, int s)                      |
 | Define a area de visao na tela.                             |
 | Os limites da tela sao:                                     |
 | e: esquerdo.                                                |
 | i: inferior.                                                |
 | d: direito.                                                 |
 | s: superior.                                                |
\*-------------------------------------------------------------*/

void tela (e, i, d, s)
int e;
int i;
int d;
int s;
{
  esqt = e;
  inft = i;
  dirt = d;
  supt = s;

  /* f e g sao auxiliares para o calculo na funcao mapeia */
  f.x = (dirt - esqt) / (dirj - esqj);
  f.y = (supt - inft) / (supj - infj);
  g.x = - (f.x * esqj) + esqt;
  g.y = - (f.y * infj) + inft;
}


/*-------------------------------------------------------------*\
 | void pega_tela (int *e, int *i, int *d, int *s)              |
 | Retorna as coordenadas da parte inferior esquerda e da      |
 | parte superior direita da tela.                             |
\*-------------------------------------------------------------*/
void pega_tela (e, i, d, s)
int *e;
int *i;
int *d;
int *s;
{
  *e = esqt;
  *i = inft;
  *d = dirt;
  *s = supt;
}


/*-------------------------------------------------------------*\
 | void clipping (float p)                                     |
 | Define a distancia minima de visao, i.e'., ate aonde a came-|
 | ra pode aproximar-se da cena.                               |
 | (AINDA NAO IMPLEMENTADO).                                   |
\*-------------------------------------------------------------*/
void clipping (float p)
{
  _perto = p;
}

float distancia_plano ()
{
  return -_perto;
}

/*-------------------------------------------------------------*\
 | void sobe_desce (float alfa)                                |
 | Sobe ou desce a posicao da camera numa trajetoria de uma    |
 | circunferencia vertical cujo centro e' o alvo.              |
\*-------------------------------------------------------------*/
void sobe_desce (float alfa)
{
  float r = distancia (&_alvo, &_posicao);
  ponto3d v0, w0,
          ut, vt, wt,
          aux;

  v0.y = cos (alfa);
  v0.z = sin (alfa);

  w0.y = -v0.z;
  w0.z = v0.y;
  
  ut.x = _obs_mundo [0][0];
  ut.y = _obs_mundo [0][1] * v0.y + _obs_mundo [0][2] * v0.z;
  ut.z = _obs_mundo [0][1] * w0.y + _obs_mundo [0][2] * w0.z;
  
  vt.x = _obs_mundo [1][0];
  vt.y = _obs_mundo [1][1] * v0.y + _obs_mundo [1][2] * v0.z;
  vt.z = _obs_mundo [1][1] * w0.y + _obs_mundo [1][2] * w0.z;

  wt.x = _obs_mundo [2][0];
  wt.y = _obs_mundo [2][1] * v0.y + _obs_mundo [2][2] * v0.z;
  wt.z = _obs_mundo [2][1] * w0.y + _obs_mundo [2][2] * w0.z;

  normalizacao(&ut, &ut);
  normalizacao(&vt, &vt);
  normalizacao(&wt, &wt);

  monta_base (&ut, &vt, &wt);
  realXvetor (r, &wt, &aux);
  adicao (&aux, &_alvo, &_posicao);
}

/*-------------------------------------------------------------*\
 | void direita_esquerda (float beta)                          |
 | Desloca-se para a esquerda ou para a direita numa trajetoria|
 | de uma circunferencia horizontal cujo centro e' o alvo.     |
\*-------------------------------------------------------------*/
void direita_esquerda (float beta)
{
  float r = distancia (&_alvo, &_posicao);
  ponto3d u0, w0,
          ut, vt,  wt,
          aux, aux_posicao;

  u0.x = cos (beta);
  u0.z = -sin (beta);

  w0.x = -u0.z;
  w0.z = u0.x;

  ut.x = _obs_mundo [0][0] * u0.x + _obs_mundo [0][2] * u0.z;
  ut.y = _obs_mundo [0][1];
  ut.z = _obs_mundo [0][0] * w0.x + _obs_mundo [0][2] * w0.z;
  
  vt.x = _obs_mundo [1][0] * u0.x + _obs_mundo [1][2] * u0.z;
  vt.y = _obs_mundo [1][1];
  vt.z = _obs_mundo [1][0] * w0.x + _obs_mundo [1][2] * w0.z;

  wt.x = _obs_mundo [2][0] * u0.x + _obs_mundo [2][2] * u0.z;
  wt.y = _obs_mundo [2][1];
  wt.z = _obs_mundo [2][0] * w0.x + _obs_mundo [2][2] * w0.z;

 
  normalizacao(&ut, &ut);
  normalizacao(&vt, &vt);
  normalizacao(&wt, &wt);

  monta_base (&ut, &vt, &wt);
  realXvetor (r, &wt, &aux);
  adicao (&aux, &_alvo, &_posicao);
  
}

/*-------------------------------------------------------------*\
 | void roda_vertical (float alfa)                             |
 | Roda a camera sobre seu proprio eixo num plano determinado  |
 | pelo vetor u(eixo x) e a posicao, ou seja, em um plano ver- |
 | tical.                                                      |
\*-------------------------------------------------------------*/
void roda_vertical (float alfa)
{
  float r = distancia (&_alvo, &_posicao);
  ponto3d v0, w0,
          ut, vt, wt,
          aux;

  v0.y = cos (alfa);
  v0.z = sin (alfa);

  w0.y = -v0.z;
  w0.z = v0.y;

  ut.x = _obs_mundo [0][0];
  ut.y = _obs_mundo [0][1] * v0.y + _obs_mundo [0][2] * v0.z;
  ut.z = _obs_mundo [0][1] * w0.y + _obs_mundo [0][2] * w0.z;
  
  vt.x = _obs_mundo [1][0];
  vt.y = _obs_mundo [1][1] * v0.y + _obs_mundo [1][2] * v0.z;
  vt.z = _obs_mundo [1][1] * w0.y + _obs_mundo [1][2] * w0.z;

  wt.x = _obs_mundo [2][0];
  wt.y = _obs_mundo [2][1] * v0.y + _obs_mundo [2][2] * v0.z;
  wt.z = _obs_mundo [2][1] * w0.y + _obs_mundo [2][2] * w0.z;

  normalizacao(&ut, &ut);
  normalizacao(&vt, &vt);
  normalizacao(&wt, &wt);

  monta_base (&ut, &vt, &wt);
  realXvetor (-r, &wt, &aux);
  adicao (&aux, &_posicao, &_alvo);
}

/*-------------------------------------------------------------*\
 | void roda_horizontal (float alfa)                           |
 | Roda a camera sobre seu proprio eixo num plano determinado  |
 | pelo vetor v(eixo y) e a posicao, ou seja, em um plano hori-|
 | zontal.                                                     |
\*-------------------------------------------------------------*/
void roda_horizontal (float alfa)
{
  float r = distancia (&_alvo, &_posicao);
  ponto3d u0, w0,
          ut, vt, wt,
          aux;

  u0.x = cos (alfa);
  u0.z = -sin (alfa);

  w0.x = -u0.z;
  w0.z = u0.x;

  ut.x = _obs_mundo [0][0] * u0.x + _obs_mundo [0][2] * u0.z;
  ut.y = _obs_mundo [0][1];
  ut.z = _obs_mundo [0][0] * w0.x + _obs_mundo [0][2] * w0.z;
  
  vt.x = _obs_mundo [1][0] * u0.x + _obs_mundo [1][2] * u0.z;
  vt.y = _obs_mundo [1][1];
  vt.z = _obs_mundo [1][0] * w0.x + _obs_mundo [1][2] * w0.z;

  wt.x = _obs_mundo [2][0] * u0.x + _obs_mundo [2][2] * u0.z;
  wt.y = _obs_mundo [2][1];
  wt.z = _obs_mundo [2][0] * w0.x + _obs_mundo [2][2] * w0.z;

  normalizacao(&ut, &ut);
  normalizacao(&vt, &vt);
  normalizacao(&wt, &wt);

  monta_base (&ut, &vt, &wt);
  realXvetor (-r, &wt, &aux);
  adicao (&aux, &_posicao, &_alvo);
}

/*
  ut.x = _obs_mundo [0][0] * u0.x +  _obs_mundo [0][1] * u0.y + _obs_mundo [0][2] * u0.z;
  ut.y = _obs_mundo [0][0] * v0.x +  _obs_mundo [0][1] * v0.y + _obs_mundo [0][2] * v0.z;
  ut.z = _obs_mundo [0][0] * w0.x +  _obs_mundo [0][1] * w0.y + _obs_mundo [0][2] * w0.z;
  
  vt.x = _obs_mundo [1][0] * u0.x +  _obs_mundo [1][1] * u0.y + _obs_mundo [1][2] * u0.z;
  vt.y = _obs_mundo [1][0] * v0.x +  _obs_mundo [1][1] * v0.y + _obs_mundo [1][2] * v0.z;
  vt.z = _obs_mundo [1][0] * w0.x +  _obs_mundo [1][1] * w0.y + _obs_mundo [1][2] * w0.z;

  wt.x = _obs_mundo [2][0] * u0.x +  _obs_mundo [2][1] * u0.y + _obs_mundo [2][2] * u0.z;
  wt.y = _obs_mundo [2][0] * v0.x +  _obs_mundo [2][1] * v0.y + _obs_mundo [2][2] * v0.z;
  wt.z = _obs_mundo [2][0] * w0.x +  _obs_mundo [2][1] * w0.y + _obs_mundo [2][2] * w0.z;
*/

/*-------------------------------------------------------------*\
 | void gira_orientacao (float gama)                           |
 | Roda a camera sobre seu proprio eixo num plano determinado  |
 | pelo vetor w(eizo z) e a posicao, ou seja, realiza uma rota-|
 | cao da camera que pode por exemplo rodar ate a cena ficar de|
 | ponta cabeca.                                               |
\*-------------------------------------------------------------*/
void gira_orientacao (float gama)
{
  ponto3d u0, v0,
          ut, vt, wt;
          
  u0.x = cos (gama);
  u0.y = sin (gama);

  v0.x = -u0.y;
  v0.y = u0.x;
      
  ut.x = _obs_mundo [0][0] * u0.x +  _obs_mundo [0][1] * u0.y;
  ut.y = _obs_mundo [0][0] * v0.x +  _obs_mundo [0][1] * v0.y;
  ut.z = _obs_mundo [0][2];
  
  vt.x = _obs_mundo [1][0] * u0.x +  _obs_mundo [1][1] * u0.y;
  vt.y = _obs_mundo [1][0] * v0.x +  _obs_mundo [1][1] * v0.y;
  vt.z = _obs_mundo [1][2];

  wt.x = _obs_mundo [2][0] * u0.x +  _obs_mundo [2][1] * u0.y;
  wt.y = _obs_mundo [2][0] * v0.x +  _obs_mundo [2][1] * v0.y;
  wt.z = _obs_mundo [2][2];
  
  normalizacao(&ut, &ut);
  normalizacao(&vt, &vt);
  normalizacao(&wt, &wt);
     
  monta_base (&ut, &vt, &wt);
}

/*-------------------------------------------------------------*\
 | void aproxima (void)                                        |
 | Aproxima a camera.                                          |
\*-------------------------------------------------------------*/
void aproxima ()
{
  ponto3d aux;

  realXvetor (_epsilon, &_direcao, &aux);
  adicao (&_posicao, &aux, &_posicao);
}

/*-------------------------------------------------------------*\
 | void afasta (void)                                          |
 | Afasta a camera.                                            |
\*-------------------------------------------------------------*/
void afasta ()
{
  ponto3d aux;

  realXvetor (_epsilon, &_direcao, &aux);
  subtracao (&_posicao, &aux, &_posicao);
}

/*-------------------------------------------------------------*\
 | void transforma_base (matriz3x3 a)                          |
 | Realiza uma transformacao na base do observador, atraves de |
 | uma matriz recebida.                                        |
 | Nota: esta matriz pode ser uma matriz arbitraria, o que pode|
 | causar distorcoes na cena.                                  |
\*-------------------------------------------------------------*/
void transforma_base (a)
matriz3x3 a;
{
  multiplicacao3x3 (_mundo_obs, a, _mundo_obs);
}

/*-------------------------------------------------------------*\
 | void calcula_base (ponto3d *VPN, ponto3d *VUP)              |
 | Calcula a base do observador atraves de dois vetores recebi-|
 | dos; VPN que e um vetor normal ao plano de projecao, e VUP  |
 | que e um vetor que indica o vetical do observador.          |
 | Retorna a base do observador.                               |
\*-------------------------------------------------------------*/
void calcula_base (VPN, VUP)
ponto3d *VPN;
ponto3d *VUP;
{
  ponto3d u, v, w;

  produto_vetorial (VPN, VUP, &u);
  normalizacao (&u, &u);

  realXvetor (-1.0, VPN, &w);
  normalizacao (&w, &w);

  produto_vetorial (&w, &u, &v);

  monta_base (&u, &v, &w);
}

/*-------------------------------------------------------------*\
 | void monta_base (ponto3d *u, ponto3d *v, ponto3d *w)        |
 | Monta a base do observador com tres vetores recebidos, os   |
 | vetores serao as colunas da matriz.                         |
\*-------------------------------------------------------------*/
void monta_base (u, v, w)
ponto3d *u;
ponto3d *v;
ponto3d *w;
{
  _mundo_obs[0][0] = u->x; _mundo_obs[0][1] = v->x; _mundo_obs[0][2] = w->x;
  _mundo_obs[1][0] = u->y; _mundo_obs[1][1] = v->y; _mundo_obs[1][2] = w->y;
  _mundo_obs[2][0] = u->z; _mundo_obs[2][1] = v->z; _mundo_obs[2][2] = w->z;

  transposta3x3 (_mundo_obs, _obs_mundo);
}

/*-------------------------------------------------------------*\
 | void muda_base (ponto3d *p, ponto3d *q)                     |
 | Muda um vetor da base do mundo para a base do observador.   |
\*-------------------------------------------------------------*/
void muda_base (p, q)
ponto3d *p;
ponto3d *q;
{
  vetorXmatriz (p, _mundo_obs, q);
}

/*-------------------------------------------------------------*\
 | void muda_referencial (ponto3d *p, ponto3d *q)              |
 | Muda um ponto da base do mundo para a base do observador.   |
\*-------------------------------------------------------------*/
void muda_referencial (p, q)
ponto3d *p;
ponto3d *q;
{
  ponto3d r;

  subtracao (p, &_posicao, &r);
  vetorXmatriz (&r, _mundo_obs, q);
}

/*-------------------------------------------------------------*\
 | void projecao_perspectiva (ponto3d *p, ponto2d *q)          |
 | Transforma um ponto de 3D para 2D usando projecao em pers-  |
 | pectiva.                                                    |
\*-------------------------------------------------------------*/
void projecao_perspectiva (p, q)
ponto3d *p;
ponto2d *q;
{
  if (p->z == 0)
    p->z = 1;
  q->x = (_foco * p->x) / p->z;
  q->y = (_foco * p->y) / p->z;
}

/*-------------------------------------------------------------*\
 | void projecao_paralela (ponto3d *p, ponto2d *q)             |
 | Transforma um ponto de 3D para 2D usando projecao paralela. |
\*-------------------------------------------------------------*/
void projecao_paralela (p, q)
ponto3d *p;
ponto2d *q;
{
  ponto3d r, s;

  subtracao (p, &_posicao, &r);
  vetorXmatriz (&r, _mundo_obs, &s);
  q->x = _foco * s.x;
  q->y = _foco * s.y;
}

/*-------------------------------------------------------------*\
 | void mapeia (ponto2d *p, ponto2d *q)                        |
 | Mapeia um ponto do plano de projecao para a tela.           |
\*-------------------------------------------------------------*/
void mapeia (p, q)
ponto2d *p;
ponto2d *q;
{
  q->x = p->x * f.x + g.x;
  q->y = p->y * f.y + g.y;
}

/* Verifica se o poligono nao e visivel */

int escondido (n,v)
ponto3d *n;
ponto3d *v;
{
   ponto3d obs_vert;
   
   subtracao(v, &_posicao, &obs_vert);

#ifdef TODASFACES
   return (0);
#else
   return (produto_escalar (&obs_vert, n) <= 0.0 ? 0 : 1);
#endif
}

